/*******************************************************************************
Copyright © 2023 Sebastian Roith

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*******************************************************************************/

(function(window) {


function isAbort({
    result      // Grund für das Ende: undefined bei manuellem Abbruch, sonst RobotSimulator.RESULT_*
}) {
    switch (result) {
        case RobotSimulator.RESULT_MISPLACED:
        case RobotSimulator.RESULT_UNDERFLOW:
        case RobotSimulator.RESULT_OVERFLOW:
        case RobotSimulator.RESULT_FULLSTACK:
            return { unlock: true };
        case RobotSimulator.RESULT_ABORT:
            return { msg: 'Schön, dass du bereits „abbruch“ kennst. Aber das gilt nicht!', accept: false };
        case RobotSimulator.RESULT_DONE:
            return { msg: "Es gab KEINE Notabschaltung", accept: false, fine: 1 };
    }
    return { fine: 1 };
};

function isOneCommand({
    code,       // In Editor eingegebener Code
    tokens      // Quellcode als Array von Token-Instanzen
}) {
    const realTokens = tokens.filter(token => (!(token instanceof TokenBracketOpen) && !(token instanceof TokenBracketClose) && !(token instanceof TokenBlockComment) && !(token instanceof TokenLineComment) && !(token instanceof TokenWhitespace )&& !(token instanceof TokenIndent)));
    let wrongTokens = realTokens.filter(token => !(token instanceof TokenCommand));
    if (wrongTokens.length)
        return { accept: false, msg: 'Schön, dass du schon mehr als nur Anweisungen kennst. Bei dieser Aufgabe sollen aber nur Algorithmen getestet werden, die aus nur einer einzigen Anweisung bestehen', line: wrongTokens[0].line };
    if (!realTokens.length)
        return { accept: false, fine: 1, msg: 'Der Methodenaufruf fehlt' };
    if (realTokens.length > 1)
        return { accept: false, fine: 1, msg: 'Verwende in dieser Aufgabe NUR EINE Anweisung', line: realTokens[1].line };
    return { accept: true };
}

const LEVELS = [
    {
        // Angezeigter Arbeitsauftrag (HTML erlaubt)
        task: 'Beim Ausführen eines Algorithmus können Probleme auftreten. Erkläre vorm Ausprobieren, wieso das hier der Fall ist:',

        // Welt und Roboterstartposition als Maschinencode
        world: '0533000000000000000100112',

        // Quelltext, der anfangs im Editor steht
        code: '# Verfügbare Anweisungen:\n#   vor         Ein Feld in Blickrichtung gehen\n#   plus        Anzahl auf aktuellem Feld um Eins erhöhen\n#   minus       Anzahl auf aktuellem Feld um Eins verringern\n#   linksrum    Um 90° nach links drehen\n\nminus',

        // Ist das Bearbeiten des Quelltextes gesperrt? (Optional, Standardwert false)
        readonly: true,

        // Muss - ggf. zusätzlich zur korrekten Fragenbeantwortung - checkCode oder checkTermination unlock=true liefern, bevor Level abgeschlossen werden kann? (Optional, Standardwert false)
        locked: false,

        // Optional: Überprüft, ob ein syntaktisch korrekter Quellcode ausgeführt werden darf; Rückgabewert true für ja, sonst { accept: "Ausführung erlauben?", line: "Zeilennummer eines Fehlers (optional)", msg: "Fehlermeldung (optional)", fine: "Als einen Fehlversuch werten? (optional)", unlock: "Level als abgeschlossen behandeln? (optional)" }
        checkCode({
            code,       // In Editor eingegebener Code
            tokens,     // Quellcode als Array von Token-Instanzen
            compiled,   // Quelltextbaum als SourceCodePart-Instanz
            worldDef    // Maschinencode der Welt zu Beginn
        }) {
            return { accept: false, msg: "Geduld! Ausprobieren darfst du das bei Aufgabe 2…" };
        },

        // Optional: Reagiert auf ein Ende der Algorithmusausführung; Rückgabewert wie checkCode
        checkTermination({
            code,       // In Editor eingegebener Code
            line,       // Zuletzt ausgeführte Zeilennummer
            result,     // Grund für das Ende: undefined bei manuellem Abbruch, sonst RobotSimulator.RESULT_*
            worldDef,   // Maschinencode der Welt zu Beginn
            worldStat   // Aktueller Status der Welt nach Algorithmusausführung
        }) {
            return {};
        },

        // Wird falsches Beantworten der Frage als Fehlversuch gewertet? (Optional, Standardwert false)
        questionWrongFine: 1,

        // Falls erst eine Frage richtig beantwortet werden muss: Welche?
        question: {
            // Optional: Titel der Frage
            title: undefined,

            // Frage
            html: "Welches Problem tritt im Beispiel oben auf?",

            // Optional: Fragentext, nach Überprüfen der Antwort
            htmlSolution: "Welches Problem tritt im Beispiel auf?",

            // Optional: Kann nur eine Antwort richtig sein?
            singlechoice: false,

            // Optional: Sollen die Antworten nicht durchgemischt werden?
            keepOrder: false,

            // Optional: Zusätzliche Erläuterung, wenn Frage falsch beantwortet wird
            response: undefined,

            // Auswählbare Antworten
            choices: [
                {
                    // Antwort
                    html: "Der Roboter kann die Anweisung nicht ausführen, weil kein Plättchen auf seinem Feld liegt",

                    // Optional: Ist das eine richtige Antwort?
                    correct: true,

                    // Optional: Anzuzeigender Text bei Fehlauswahl
                    response: undefined
                },
                { html: "Der Algorithmus ist nicht eindeutig", response: 'Doch: Der Algorithmus beschreibt eindeutig, dass er vom aktuellen Feld ein Plättchen entfernen soll' },
                { html: "Der Algorithmus entspricht nicht den Vorgaben der verwendeten Programmiersprache", response: 'Doch: Der Aufbau ist einwandfrei' },
                { html: "Die Klammern nach dem Methodennamen <code>minus</code> fehlen", response: "Nein, die Klammern sind optional" }
            ]
        }
    },
    {
        task: 'Probiere nun aus, was passiert, wenn du den Algorithmus dennoch startest!',
        world: '0533000000000000000100112',
        code: '# Verfügbare Anweisungen:\n#   vor         Ein Feld in Blickrichtung gehen\n#   plus        Anzahl auf aktuellem Feld um Eins erhöhen\n#   minus       Anzahl auf aktuellem Feld um Eins verringern\n#   linksrum    Um 90° nach links drehen\n\nminus',
        readonly: true,
        locked: true,
        checkTermination: isAbort,
        questionWrongFine: 1,
        question: {
            html: "Was passiert, wenn die Anweisung nicht ausgeführt werden kann?",
            choices: [
                { html: "Es erscheinen zusätzliche Plättchen" },
                { html: "Der Roboter wechselt auf ein anderes Feld" },
                { html: "Der Roboter führt die Anweisung nicht aus", correct: true },
                { html: "Die Ausführung des Algorithmus pausiert, bis genug Plättchen vorhanden sind", response: "An der Meldung über dem Eingabefeld und weil du wieder auf „Roboter starten“ klicken kannst, erkennst du, dass der Algorithmus nicht pausiert." },
                { html: "Statt einer Erfolgsmeldung gibt es eine Fehlermeldung", correct: true, response: "Die Erfolgs- bzw. Fehlermeldung findest du über dem Eingabefeld für den Algorithmus." },
                { html: "Der Grund für die Notabschaltung wird angezeigt.", correct: true, response: "Der Grund für die Notabschaltung steht in der Fehlermeldung über dem Eingabefeld für den Algorithmus." }
            ]
        }
    },
    {
        task: '<p>Zum Schutz vor Beschädigung schaltet sich der Roboter sofort aus, wenn eine Anweisung nicht durchführbar ist. Der Algorithmus wird also abgebrochen. Solch eine <strong>Notabschaltung</strong> kann verschiedene Gründe haben:</p><ul><li><code>minus</code> soll auf einem leeren Feld ausgeführt werden</li><li><code>plus</code> soll auf einem Feld ausgeführt werden, auf dem bereits die Maximalanzahl an Plättchen liegt: 15</li><li><code>vor</code> soll ausgeführt werden, obwohl der Roboter kein begehbares Feld vor sich (in Blickrichtung) hat.</li></ul><p>Gelingt es dir, hier auch für eine Notabschaltung zu sorgen?</p>',
        world: '05110F002',
        code: '# Verfügbare Anweisungen:\n#   vor         Ein Feld in Blickrichtung gehen\n#   plus        Anzahl auf aktuellem Feld um Eins erhöhen\n#   minus       Anzahl auf aktuellem Feld um Eins verringern\n#   linksrum    Um 90° nach links drehen\n\nminus',
        locked: true,
        checkCode: isOneCommand,
        checkTermination: isAbort,
        questionWrongFine: 1,
        question: {
            html: "Bei welchen der dir bekannten Anweisungen ist das Auftreten einer Notabschaltung allgemein möglich?",
            choices: [
                { html: "Anweisung <code>vor</code>", correct: true },
                { html: "Anweisung <code>plus</code>", correct: true },
                { html: "Anweisung <code>minus</code>", correct: true },
                { html: "Anweisung <code>linksrum</code>" },
                { html: "Anweisung <code>zurueck</code>", response: "Diese Anweisung gibt es gar nicht!" }
            ]
        }
    },
    {
        task: 'Schreibe einen Algorithmus mit nur einer einzigen Anweisung, der in der angezeigten Welt zu einer Notabschaltung führt:',
        world: '0533010000011010010000011',
        code: '# Verfügbare Anweisungen:\n#   vor         Ein Feld in Blickrichtung gehen\n#   plus        Anzahl auf aktuellem Feld um Eins erhöhen\n#   minus       Anzahl auf aktuellem Feld um Eins verringern\n#   linksrum    Um 90° nach links drehen\n\nminus',
        locked: true,
        checkCode: isOneCommand,
        checkTermination: isAbort
    },
    {
        task: 'Schreibe einen Algorithmus mit nur einer einzigen Anweisung, der in der nun angezeigten Welt zu einer Notabschaltung führt:',
        world: '0533010000001010010000010',
        code: '# Verfügbare Anweisungen:\n#   vor         Ein Feld in Blickrichtung gehen\n#   plus        Anzahl auf aktuellem Feld um Eins erhöhen\n#   minus       Anzahl auf aktuellem Feld um Eins verringern\n#   linksrum    Um 90° nach links drehen\n\nvor',
        locked: true,
        checkCode: isOneCommand,
        checkTermination: isAbort
    },
    {
        task: 'Schreibe einen Algorithmus mit nur einer einzigen Anweisung, der in der neuen angezeigten Welt zu einer Notabschaltung führt:',
        world: '0533010000011010010000013',
        code: '# Verfügbare Anweisungen:\n#   vor         Ein Feld in Blickrichtung gehen\n#   plus        Anzahl auf aktuellem Feld um Eins erhöhen\n#   minus       Anzahl auf aktuellem Feld um Eins verringern\n#   linksrum    Um 90° nach links drehen\n\nminus',
        locked: true,
        checkCode: isOneCommand,
        checkTermination: isAbort
    },
    {
        task: 'Schreibe einen Algorithmus mit nur einer einzigen Anweisung, der in dieser angezeigten Welt zu einer Notabschaltung führt:',
        world: '05330100000F1010010000012',
        code: '# Verfügbare Anweisungen:\n#   vor         Ein Feld in Blickrichtung gehen\n#   plus        Anzahl auf aktuellem Feld um Eins erhöhen\n#   minus       Anzahl auf aktuellem Feld um Eins verringern\n#   linksrum    Um 90° nach links drehen\n\nvor',
        locked: true,
        checkCode: isOneCommand,
        checkTermination: isAbort
    },
    {
        task: 'Du kannst eine Notabschaltung auch immer durch die Anweisung <code>abbruch</code> erzwingen. Probiere das aus:',
        world: '0533010203040506070809112',
        code: '# Verfügbare Anweisungen:\n#   vor         Ein Feld in Blickrichtung gehen\n#   plus        Anzahl auf aktuellem Feld um Eins erhöhen\n#   minus       Anzahl auf aktuellem Feld um Eins verringern\n#   linksrum    Um 90° nach links drehen\n#   abbruch     Sofortige Notabschaltung\n\nplus',
        locked: true,
        checkCode: isOneCommand,
        checkTermination(code, line, result) { return (result = RobotSimulator.RESULT_ABORT) ? { unlock: true } : { fine: 1 } }
    },
    {
        task: 'Schreibe und teste zum Abschluss noch einen Algorithmus, der zu <strong>keiner</strong> Notabschaltung führt:',
        world: '051100000',
        code: '# Verfügbare Anweisungen:\n#   vor         Ein Feld in Blickrichtung gehen\n#   plus        Anzahl auf aktuellem Feld um Eins erhöhen\n#   minus       Anzahl auf aktuellem Feld um Eins verringern\n#   linksrum    Um 90° nach links drehen\n#   abbruch     Sofortige Notabschaltung\n\nabbruch',
        locked: true,
        checkCode: isOneCommand,
        checkTermination({ result }) { return (result == RobotSimulator.RESULT_DONE) ? { unlock: true } : { fine: 1 } }
    }
];

window.roboterCodingTask = {
    levels: LEVELS
};
})(window);
